require 'omniauth/strategies/oauth2'
require 'jwt'

module OmniAuth
  module Strategies
    class PracticeFusion < OmniAuth::Strategies::OAuth2

      option :name, 'practicefusion'

      option :authorize_options, [:response_type, :client_id, :redirect_uri, :Fauthcode, :scope]
      option :client_options, {
        site: 'https://api.practicefusion.com',
        authorize_url: '/ehr/oauth2/auth',
        token_url: '/ehr/oauth2/token'
      }

      uid { raw_info['ehrUserGuid'] }

      info do
        prune!({
          title: value_or_blank(raw_info['title']),
          first_name: value_or_blank(raw_info['firstName']),
          last_name: value_or_blank(raw_info['lastName']),
          provider_guid: value_or_blank(raw_info['providerGuid']),
          practice_guid: value_or_blank(raw_info['practiceGuid']),
          email: value_or_blank(raw_info['loginEmailAddress'])
        })
      end

      extra do
        hash = {}
        hash[:raw_info] = raw_info
        prune! hash
      end

      credentials do
        # these are copy-paste from OmniAuth::Strategy::OAuth2
        hash = {'token' => access_token.token}
        hash.merge!('refresh_token' => access_token.refresh_token) if access_token.expires? && access_token.refresh_token
        hash.merge!('expires_at' => access_token.expires_at) if access_token.expires?
        hash.merge!('expires' => access_token.expires?)

        # additional parameters returned by Practice Fusion
        hash.merge!('token_type' => access_token['token_type'])
        hash.merge!('scope' => access_token['scope'])
        hash.merge!('resource_owner' => access_token['resource_owner'])
        hash.merge!('resource_owner_authority' => access_token['resource_owner_authority'])
        hash.merge!('pf_session_token' => access_token['pf_session_token']) if access_token.params.key?('pf_session_token')
        hash
      end

      # override callback_url so we don't include query parameters
      # See https://github.com/intridea/omniauth-oauth2/issues/81 for more details
      def callback_url
        full_host + script_name + callback_path
      end

      def raw_info
        @raw_info ||= access_token.get('/ehr/v1/users/me').parsed
      end

      # add the 'authCookie' parameter to the outbound request
      # the 'state' parameter becomes an encrypted JWT of the session
      def authorize_params
        super.merge(session['omniauth.params'].slice('authCookie')).tap do |authz_params|
          # so omniauth decided (in their infinite wisdom) that instead of setting the
          # omniauth.origin parameter using their session method, they were gonna access
          # env['rack.session'] directly, so we need to pull that out into the session
          session['omniauth.origin'] = env['rack.session']['omniauth.origin']
          authz_params[:state] = JWT.encode(session, options.client_secret)
        end
      end

      # override the callback_phase so we can validate the state token signature
      def callback_phase
        # If decode_state_parameter returns the correct state token, then
        #   request.params['state'] == session['omniauth.state']
        # which passes the CSRF check of omniauth-oauth2

        # If decode_state_parameter returns nil, then this is true:
        #   request.params['state'].to_s.empty?
        # which fails the CSRF check of omniauth-oauth2
        request.params['state'] = decode_state_parameter

        # omniauth sets the environment up _before_ the callback_phase
        # begins, but we need to set it up after we've decoded the session
        @env['omniauth.origin'] = session.delete('omniauth.origin')
        @env['omniauth.origin'] = nil if env['omniauth.origin'] == ''
        @env['omniauth.params'] = session.delete('omniauth.params')

        # And now back to our regularly scheduled program
        super
      end

      # instead of using the session store, just collect into a hash
      def session
        @fake_session ||= {}
      end

      def session=(obj)
        @fake_session = obj
      end

      private

      def prune!(hash)
        hash.delete_if do |_, v|
          prune!(v) if v.is_a?(Hash)
          v.nil? || (v.respond_to?(:empty?) && v.empty?)
        end
      end

      def value_or_blank(value)
        (value.blank?) ? '' : value
      end

      def decode_state_parameter
        # this will throw an exception if the state parameter is not valid ciphertext
        # otherwise, decode the JWT and stick the object back into our fake session
        self.session, _jwt = JWT.decode(request.params["state"].to_s, options.client_secret)
        session['omniauth.state']
      rescue JWT::VerificationError, JWT::DecodeError
        nil
      end
    end
  end
end
